"""Factory for serializing messages according to WolkAbout Protocol."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
from time import time
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk import logger_factory
from wolk.interface.message_factory import MessageFactory
from wolk.model.actuator_status import ActuatorStatus
from wolk.model.alarm import Alarm
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_management_status_type import FileManagementStatusType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType
from wolk.model.message import Message
from wolk.model.sensor_reading import SensorReading


class WolkAboutProtocolMessageFactory(MessageFactory):
    """Serialize messages to be sent to WolkAbout IoT Platform."""

    DEVICE_PATH_PREFIX = "d/"
    REFERENCE_PATH_PREFIX = "r/"
    CHANNEL_WILDCARD = "#"
    CHANNEL_DELIMITER = "/"
    LAST_WILL = "lastwill/"
    SENSOR_READING = "d2p/sensor_reading/"
    ALARM = "d2p/events/"
    ACTUATOR_STATUS = "d2p/actuator_status/"
    CONFIGURATION_STATUS = "d2p/configuration_get/"
    FILE_BINARY_REQUEST = "d2p/file_binary_request/"
    FIRMWARE_VERSION_UPDATE = "d2p/firmware_version_update/"
    FIRMWARE_UPDATE_STATUS = "d2p/firmware_update_status/"
    FILE_LIST_UPDATE = "d2p/file_list_update/"
    FILE_LIST_RESPONSE = "d2p/file_list_response/"
    FILE_UPLOAD_STATUS = "d2p/file_upload_status/"
    FILE_URL_DOWNLOAD_STATUS = "d2p/file_url_download_status/"
    KEEP_ALIVE = "ping/"

    def __init__(self, device_key: str) -> None:
        """
        Create a factory for serializing messages.

        :param device_key: Device key to use when serializing messages
        :type device_key: str
        """
        self.device_key = device_key
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(f"Device key: {device_key}")

    def make_last_will_message(self) -> Message:
        """
        Serialize a last will message if device disconnects unexpectedly.

        :returns: Last will message
        :rtype: Message
        """
        topic = self.LAST_WILL + self.device_key
        message = Message(topic)
        self.logger.debug(f"{message}")

        return message

    def make_keep_alive_message(self) -> Message:
        """
        Serialize a keep alive message.

        :returns: keep alive message
        :rtype: Message
        """
        topic = self.KEEP_ALIVE + self.device_key
        message = Message(topic)
        self.logger.debug(f"{message}")

        return message

    def make_from_sensor_reading(
        self, reading: Union[SensorReading, List[SensorReading]]
    ) -> Message:
        """
        Serialize the sensor reading to be sent to WolkAbout IoT Platform.

        :param reading: Sensor reading to be serialized
        :type reading: Union[SensorReading, List[SensorReading]]
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{reading}")

        if isinstance(reading, SensorReading):
            reference = reading.reference
        else:
            reference = reading[0].reference

        topic = (
            self.SENSOR_READING
            + self.DEVICE_PATH_PREFIX
            + self.device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + reference
        )

        if not isinstance(reading, list):
            reading = [reading]

        payload: List[Dict[str, Union[int, str]]] = []

        for single_reading in reading:
            serialized_reading: Dict[str, Union[int, str]] = {}
            if isinstance(single_reading.value, tuple):
                serialized_reading["data"] = ",".join(
                    map(str, single_reading.value)
                )

            elif isinstance(single_reading.value, bool):
                serialized_reading["data"] = str(single_reading.value).lower()

            else:
                serialized_reading["data"] = str(single_reading.value)

            if single_reading.timestamp is not None:
                serialized_reading["utc"] = int(single_reading.timestamp)
            else:
                serialized_reading["utc"] = round(time() * 1000)

            payload.append(serialized_reading)

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_sensor_readings(
        self,
        readings: Dict[
            str,
            Union[bool, str, int, Tuple[int, ...], float, Tuple[float, ...]],
        ],
        timestamp: Optional[int] = None,
    ) -> Message:
        """
        Serialize multiple sensor readings to send to WolkAbout IoT Platform.

        :param readings: Dictionary of reference: value pairs
        :type readings: Dict[str,Union[bool, str, int, Tuple[int, ...], float, Tuple[float, ...]]]
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"readings:{readings}, timestamp:{timestamp}")

        topic = self.SENSOR_READING + self.DEVICE_PATH_PREFIX + self.device_key
        payload = readings
        for reference, value in payload.items():
            if isinstance(value, tuple):
                payload[reference] = ",".join(map(str, value))

            elif isinstance(value, bool):
                payload[reference] = str(value).lower()
            else:
                payload[reference] = str(value)

        if timestamp is not None:
            payload.update({"utc": int(timestamp)})
        else:
            payload.update({"utc": round(time()) * 1000})

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_alarm(self, alarm: Alarm) -> Message:
        """
        Serialize the alarm to be sent to WolkAbout IoT Platform.

        :param alarm: Alarm event to be serialized
        :type alarm: Alarm
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{alarm}")

        topic = (
            self.ALARM
            + self.DEVICE_PATH_PREFIX
            + self.device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + alarm.reference
        )

        payload: Dict[str, Union[str, int]] = {
            "active": str(alarm.active).lower()
        }

        if alarm.timestamp is not None:
            payload["utc"] = alarm.timestamp
        else:
            payload["utc"] = round(time()) * 1000

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_actuator_status(self, actuator: ActuatorStatus) -> Message:
        """
        Serialize the actuator status to be sent to WolkAbout IoT Platform.

        :param actuator: Actuator status to be serialized
        :type actuator: ActuatorStatus
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{actuator}")

        topic = (
            self.ACTUATOR_STATUS
            + self.DEVICE_PATH_PREFIX
            + self.device_key
            + self.CHANNEL_DELIMITER
            + self.REFERENCE_PATH_PREFIX
            + actuator.reference
        )

        if isinstance(actuator.value, bool):
            actuator.value = str(actuator.value).lower()

        payload = {"status": actuator.state.value}
        payload["value"] = str(actuator.value)

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_configuration(
        self, configuration: Dict[str, Optional[Union[int, float, bool, str]]]
    ) -> Message:
        """
        Report device's configuration to WolkAbout IoT Platform.

        :param configuration: Device's current configuration
        :type configuration: dict
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{configuration}")
        topic = (
            self.CONFIGURATION_STATUS
            + self.DEVICE_PATH_PREFIX
            + self.device_key
        )
        payload: Dict[str, Dict[str, str]] = {"values": {}}

        for reference, value in configuration.items():
            if isinstance(value, bool):
                value = str(value).lower()
            else:
                value = str(value)

            payload["values"].update({reference: value})

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_package_request(
        self, file_name: str, chunk_index: int, chunk_size: int
    ) -> Message:
        """
        Request a package of the file from WolkAbout IoT Platform.

        :param file_name: Name of the file that contains the requested package
        :type file_name: str
        :param chunk_index: Index of the requested package
        :type chunk_index: int
        :param chunk_size: Size of the requested package
        :type chunk_size: int
        :returns: message
        :rtype: Message
        """
        self.logger.debug(
            f"file_name: '{file_name}', "
            f"chunk_index: {chunk_index}, "
            f"chunk_size: {chunk_size}"
        )
        topic = (
            self.FILE_BINARY_REQUEST
            + self.DEVICE_PATH_PREFIX
            + self.device_key
        )
        payload = {
            "fileName": file_name,
            "chunkIndex": chunk_index,
            "chunkSize": chunk_size,
        }
        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_file_list_update(self, file_list: List[str]) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[str]
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{file_list}")
        topic = (
            self.FILE_LIST_UPDATE + self.DEVICE_PATH_PREFIX + self.device_key
        )
        message = Message(
            topic, json.dumps([{"fileName": file} for file in file_list])
        )
        self.logger.debug(f"{message}")

        return message

    def make_from_file_list_request(self, file_list: List[str]) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[str]
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{file_list}")
        topic = (
            self.FILE_LIST_RESPONSE + self.DEVICE_PATH_PREFIX + self.device_key
        )
        message = Message(
            topic, json.dumps([{"fileName": file} for file in file_list])
        )
        self.logger.debug(f"{message}")

        return message

    def make_from_file_management_status(
        self, status: FileManagementStatus, file_name: str
    ) -> Message:
        """
        Serialize device's current file management status.

        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Name of file being transfered
        :type file_name: str
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f" status: {status}, file_name: {file_name}")
        topic = (
            self.FILE_UPLOAD_STATUS + self.DEVICE_PATH_PREFIX + self.device_key
        )
        payload = {"fileName": file_name, "status": status.status.value}
        if (
            status.status == FileManagementStatusType.ERROR
            and status.error is not None
        ):
            payload["error"] = status.error.value

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> Message:
        """
        Serialize device's current file URL download status.

        :param file_url: URL from where the file is to be downloaded
        :type file_url: str
        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Only present when download of file is completed
        :type file_name: Optional[str]
        """
        self.logger.debug(
            f"file_url: {file_url}, status: {status}, file_name: {file_name}"
        )
        topic = (
            self.FILE_URL_DOWNLOAD_STATUS
            + self.DEVICE_PATH_PREFIX
            + self.device_key
        )
        payload = {"fileUrl": file_url, "status": status.status.value}

        if (
            status.status == FileManagementStatusType.ERROR
            and status.error is not None
        ):
            payload["error"] = status.error.value

        if file_name is not None:
            payload["fileName"] = file_name

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_firmware_update_status(
        self, firmware_update_status: FirmwareUpdateStatus
    ) -> Message:
        """
        Serialize firmware update status to be sent to WolkAbout IoT Platform.

        :param firmware_update_status: Firmware update status to be serialized
        :type firmware_status: FirmwareUpdateStatus
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"{firmware_update_status}")
        topic = (
            self.FIRMWARE_UPDATE_STATUS
            + self.DEVICE_PATH_PREFIX
            + self.device_key
        )
        payload = {"status": firmware_update_status.status.value}

        if (
            firmware_update_status.status == FirmwareUpdateStatusType.ERROR
            and firmware_update_status.error is not None
        ):
            payload["error"] = firmware_update_status.error.value

        message = Message(topic, json.dumps(payload))
        self.logger.debug(f"{message}")

        return message

    def make_from_firmware_version_update(self, version: str) -> Message:
        """
        Report the current version of firmware to WolkAbout IoT Platform.

        :param version: Firmware version to report
        :type version: str
        :returns: message
        :rtype: Message
        """
        self.logger.debug(f"version: {version}")
        topic = (
            self.FIRMWARE_VERSION_UPDATE
            + self.DEVICE_PATH_PREFIX
            + self.device_key
        )
        message = Message(topic, str(version))
        self.logger.debug(f"{message}")

        return message



"""Enables firmware update for device."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from threading import Timer
from typing import Callable
from typing import Optional

from wolk import logger_factory
from wolk.interface.firmware_handler import FirmwareHandler
from wolk.interface.firmware_update import FirmwareUpdate
from wolk.model.firmware_update_error_type import FirmwareUpdateErrorType
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.firmware_update_status_type import FirmwareUpdateStatusType


class OSFirmwareUpdate(FirmwareUpdate):
    """Responsible for everything related to the firmware update process."""

    def __init__(
        self,
        firmware_handler: FirmwareHandler,
        status_callback: Callable[[FirmwareUpdateStatus], None],
    ) -> None:
        """
        Enable firmware update for device.

        :param firmware_handler: Install firmware and provide current version
        :type firmware_handler: FirmwareHandler
        :param status_callback: Reporting firmware update status
        :type status_callback: Callable[[FirmwareUpdateStatus], None]
        """
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(f"firmware_handler: {firmware_handler}")
        self.status_callback = status_callback
        if not isinstance(firmware_handler, FirmwareHandler):
            raise ValueError(
                f"Given firmware handler {firmware_handler} is not "
                "an instance of FirmwareHandler"
            )
        self.firmware_handler = firmware_handler
        self.current_status: Optional[FirmwareUpdateStatus] = None
        self.install_timer: Optional[Timer] = None

    def get_current_version(self) -> str:
        """
        Return device's current firmware version.

        :returns: Firmware version
        :rtype: str
        """
        return self.firmware_handler.get_current_version()

    def handle_install(self, file_path: str) -> None:
        """
        Handle received firmware installation command.

        :param file_path: Firmware file to install
        :type file_path: str
        """
        self.logger.debug(
            f"Handling install command with path path: {file_path}"
        )

        if self.current_status is not None:

            self.logger.warning(
                "Not in idle status, ignoring install command."
            )
            return

        if os.path.exists("last_firmware_version.txt"):
            self.current_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ERROR,
                FirmwareUpdateErrorType.UNSPECIFIED_ERROR,
            )
            self.logger.error("Previous firmware update did not complete!")
            self.status_callback(self.current_status)
            self._reset_state()
            return

        if not os.path.exists(file_path):
            self.current_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ERROR,
                FirmwareUpdateErrorType.FILE_NOT_PRESENT,
            )
            self.logger.error("File not present at given path!")
            self.status_callback(self.current_status)
            self._reset_state()
            return

        with open("last_firmware_version.txt", "w") as file:
            file.write(self.firmware_handler.get_current_version())

        self.current_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.INSTALLING
        )
        self.logger.info(
            "Beginning firmware installation process "
            f"with file path: {file_path}"
        )
        self.status_callback(self.current_status)

        self.install_timer = Timer(
            5.0,
            self.firmware_handler.install_firmware(file_path),  # type: ignore
        )  # For possible abort command
        self.install_timer.start()

    def handle_abort(self) -> None:
        """Handle the abort command received from the platform."""
        if self.install_timer:
            self.logger.info("Stopping installation timer")
            self.install_timer.cancel()

        if self.current_status is not None:

            self.logger.info("Aborting firmware installation")
            self.current_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ABORTED
            )
            self.status_callback(self.current_status)
            self._reset_state()
            if os.path.exists("last_firmware_version.txt"):
                os.remove("last_firmware_version.txt")

    def report_result(self) -> None:
        """Report the result of the firmware installation process."""
        self.logger.debug("Reporting firmware update result")

        if not os.path.exists("last_firmware_version.txt"):
            self.logger.debug("No stored firmware version found")
            return

        with open("last_firmware_version.txt") as file:
            last_firmware_version = file.read()

        if (
            last_firmware_version
            == self.firmware_handler.get_current_version()
        ):
            self.logger.warning(
                "Firmware version unchanged, reporting installation failed"
            )
            self.current_status = FirmwareUpdateStatus(
                FirmwareUpdateStatusType.ERROR,
                FirmwareUpdateErrorType.INSTALLATION_FAILED,
            )
            self.status_callback(self.current_status)
            self._reset_state()
            os.remove("last_firmware_version.txt")
            return

        self.logger.info(
            "Firmware version changed, reporting installation completed"
        )
        self.current_status = FirmwareUpdateStatus(
            FirmwareUpdateStatusType.COMPLETED
        )
        self.status_callback(self.current_status)
        self._reset_state()
        os.remove("last_firmware_version.txt")

    def _reset_state(self) -> None:
        """Reset the state of the firmware update process."""
        self.current_status = None
        self.install_timer = None



"""Deserialize messages received in WolkAbout Protocol format."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
from distutils.util import strtobool
from typing import List
from typing import Tuple

from wolk import logger_factory
from wolk.interface.message_deserializer import MessageDeserializer
from wolk.model.actuator_command import ActuatorCommand
from wolk.model.configuration_command import ConfigurationCommand
from wolk.model.device import Device
from wolk.model.file_transfer_package import FileTransferPackage
from wolk.model.message import Message


class WolkAboutProtocolMessageDeserializer(MessageDeserializer):
    """
    Deserialize messages received from the WolkAbout IoT Platform.

    :ivar logger: Logger instance issued by wolk.LoggerFactory
    :vartype logger: logging.Logger
    """

    DEVICE_PATH_DELIMITER = "d/"
    REFERENCE_PATH_PREFIX = "r/"
    CHANNEL_DELIMITER = "/"
    KEEP_ALIVE_RESPONSE = "pong/"
    ACTUATOR_SET = "p2d/actuator_set/"
    CONFIGURATION_SET = "p2d/configuration_set/"
    FILE_BINARY_RESPONSE = "p2d/file_binary_response/"
    FILE_DELETE = "p2d/file_delete/"
    FILE_PURGE = "p2d/file_purge/"
    FILE_LIST_CONFIRM = "p2d/file_list_confirm/"
    FILE_LIST_REQUEST = "p2d/file_list_request/"
    FILE_UPLOAD_ABORT = "p2d/file_upload_abort/"
    FILE_UPLOAD_INITIATE = "p2d/file_upload_initiate/"
    FILE_URL_DOWNLOAD_ABORT = "p2d/file_url_download_abort/"
    FILE_URL_DOWNLOAD_INITIATE = "p2d/file_url_download_initiate/"
    FIRMWARE_UPDATE_ABORT = "p2d/firmware_update_abort/"
    FIRMWARE_UPDATE_INSTALL = "p2d/firmware_update_install/"

    def __init__(self, device: Device) -> None:
        """
        Create inbound topics from device key.

        :param device: Device key and actuator references for inbound topics
        :type device: Device
        """
        self.logger = logger_factory.logger_factory.get_logger(
            str(self.__class__.__name__)
        )
        self.logger.debug(f"{device}")

        self.inbound_topics = [
            self.KEEP_ALIVE_RESPONSE + device.key,
            self.CONFIGURATION_SET + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_BINARY_RESPONSE
            + self.DEVICE_PATH_DELIMITER
            + device.key,
            self.FILE_DELETE + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_PURGE + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_LIST_CONFIRM + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_LIST_REQUEST + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_UPLOAD_ABORT + self.DEVICE_PATH_DELIMITER + device.key,
            self.FILE_UPLOAD_INITIATE
            + self.DEVICE_PATH_DELIMITER
            + device.key,
            self.FILE_URL_DOWNLOAD_ABORT
            + self.DEVICE_PATH_DELIMITER
            + device.key,
            self.FILE_URL_DOWNLOAD_INITIATE
            + self.DEVICE_PATH_DELIMITER
            + device.key,
            self.FIRMWARE_UPDATE_ABORT
            + self.DEVICE_PATH_DELIMITER
            + device.key,
            self.FIRMWARE_UPDATE_INSTALL
            + self.DEVICE_PATH_DELIMITER
            + device.key,
        ]

        for reference in device.actuator_references:
            self.inbound_topics.append(
                self.ACTUATOR_SET
                + self.DEVICE_PATH_DELIMITER
                + device.key
                + self.CHANNEL_DELIMITER
                + self.REFERENCE_PATH_PREFIX
                + reference
            )
        self.logger.debug(f"inbound topics: {self.inbound_topics}")

    def get_inbound_topics(self) -> List[str]:
        """
        Return list of inbound topics for device.

        :returns: List of topics to subscribe to
        :rtype: List[str]
        """
        return self.inbound_topics

    def is_keep_alive_response(self, message: Message) -> bool:
        """
        Check if message is keep alive response.

        :param message: The message received
        :type message: Message
        :returns: keep_alive_response
        :rtype: bool
        """
        keep_alive_response = message.topic.startswith(
            self.KEEP_ALIVE_RESPONSE
        )
        self.logger.debug(
            f"{message.topic} is keep alive response: {keep_alive_response}"
        )
        return keep_alive_response

    def is_actuation_command(self, message: Message) -> bool:
        """
        Check if message is actuation command.

        :param message: The message received
        :type message: Message
        :returns: actuation_command
        :rtype: bool
        """
        actuation_command = message.topic.startswith(self.ACTUATOR_SET)
        self.logger.debug(
            f"{message.topic} is actuation command: {actuation_command}"
        )
        return actuation_command

    def is_firmware_install(self, message: Message) -> bool:
        """
        Check if message is firmware update install command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_install
        :rtype: bool
        """
        firmware_update_install = message.topic.startswith(
            self.FIRMWARE_UPDATE_INSTALL
        )
        self.logger.debug(
            f"{message.topic} is firmware install: {firmware_update_install}"
        )
        return firmware_update_install

    def is_firmware_abort(self, message: Message) -> bool:
        """
        Check if message is firmware update command.

        :param message: The message received
        :type message: Message
        :returns: firmware_update_abort
        :rtype: bool
        """
        firmware_update_abort = message.topic.startswith(
            self.FIRMWARE_UPDATE_ABORT
        )
        self.logger.debug(
            f"{message.topic} is firmware abort: {firmware_update_abort}"
        )
        return firmware_update_abort

    def is_file_binary_response(self, message: Message) -> bool:
        """
        Check if message is file binary message.

        :param message: The message received
        :type message: Message
        :returns: file_binary
        :rtype: bool
        """
        file_binary = message.topic.startswith(self.FILE_BINARY_RESPONSE)
        self.logger.debug(f"{message.topic} is file binary: {file_binary}")
        return file_binary

    def is_configuration_command(self, message: Message) -> bool:
        """
        Check if message is configuration command.

        :param message: The message received
        :type message: Message
        :returns: configuration
        :rtype: bool
        """
        configuration = message.topic.startswith(self.CONFIGURATION_SET)
        self.logger.debug(f"{message.topic} is configuration: {configuration}")
        return configuration

    def is_file_delete_command(self, message: Message) -> bool:
        """
        Check if message if file delete command.

        :param message: The message received
        :type message: Message
        :returns: file_delete_command
        :rtype: bool
        """
        file_delete_command = message.topic.startswith(self.FILE_DELETE)
        self.logger.debug(
            f"{message.topic} is file delete: {file_delete_command}"
        )
        return file_delete_command

    def is_file_purge_command(self, message: Message) -> bool:
        """
        Check if message if file purge command.

        :param message: The message received
        :type message: Message
        :returns: file_purge_command
        :rtype: bool
        """
        file_purge_command = message.topic.startswith(self.FILE_PURGE)
        self.logger.debug(
            f"{message.topic} is file purge: {file_purge_command}"
        )
        return file_purge_command

    def is_file_list_confirm(self, message: Message) -> bool:
        """
        Check if message is file list confirm.

        :param message: The message received
        :type message: Message
        :returns: file_list_confirm
        :rtype: bool
        """
        file_list_confirm = message.topic.startswith(self.FILE_LIST_CONFIRM)
        self.logger.debug(
            f"{message.topic} is file list confirm: {file_list_confirm}"
        )
        return file_list_confirm

    def is_file_list_request(self, message: Message) -> bool:
        """
        Check if message is file list request.

        :param message: The message received
        :type message: Message
        :returns: file_list_request
        :rtype: bool
        """
        file_list_request = message.topic.startswith(self.FILE_LIST_REQUEST)
        self.logger.debug(
            f"{message.topic} is file list request: {file_list_request}"
        )
        return file_list_request

    def is_file_upload_initiate(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_initiate
        :rtype: bool
        """
        file_upload_initiate = message.topic.startswith(
            self.FILE_UPLOAD_INITIATE
        )
        self.logger.debug(
            f"{message.topic} is file upload init: {file_upload_initiate}"
        )
        return file_upload_initiate

    def is_file_upload_abort(self, message: Message) -> bool:
        """
        Check if message is file upload command.

        :param message: The message received
        :type message: Message
        :returns: file_upload_abort_command
        :rtype: bool
        """
        file_upload_abort_command = message.topic.startswith(
            self.FILE_UPLOAD_ABORT
        )
        self.logger.debug(
            f"{message.topic} is file purge: {file_upload_abort_command}"
        )
        return file_upload_abort_command

    def is_file_url_initiate(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_init
        :rtype: bool
        """
        file_url_download_init = message.topic.startswith(
            self.FILE_URL_DOWNLOAD_INITIATE
        )
        self.logger.debug(
            f"{message.topic} is file url download: {file_url_download_init}"
        )
        return file_url_download_init

    def is_file_url_abort(self, message: Message) -> bool:
        """
        Check if message is file URL download command.

        :param message: The message received
        :type message: Message
        :returns: file_url_download_abort
        :rtype: bool
        """
        file_url_download_abort = message.topic.startswith(
            self.FILE_URL_DOWNLOAD_ABORT
        )
        self.logger.debug(
            f"{message.topic} is file url abort: {file_url_download_abort}"
        )
        return file_url_download_abort

    def parse_keep_alive_response(self, message: Message) -> int:
        """
        Parse the message into an UTC timestamp.

        :param message: The message received
        :type message: Message
        :returns: timestamp
        :rtype: int
        """
        self.logger.debug(f"{message}")
        payload = json.loads(message.payload.decode("utf-8"))  # type: ignore

        timestamp = payload["value"]

        self.logger.debug(f"received timestamp: {timestamp}")
        return timestamp

    def parse_actuator_command(self, message: Message) -> ActuatorCommand:
        """
        Parse the message into an actuation command.

        :param message: Message to be deserialized
        :type message: Message

        :returns: actuation
        :rtype: ActuatorCommand
        """
        self.logger.debug(f"{message}")
        reference = message.topic.split("/")[-1]

        payload = json.loads(message.payload.decode("utf-8"))  # type: ignore
        value = payload["value"]
        if "\\n" in value:
            value = value.replace("\\n", "\n")
        if value in ["true", "false"]:
            value = bool(strtobool(value))
        else:
            try:
                value = float(value)
            except ValueError:
                try:
                    value = int(value)
                except ValueError:
                    pass

        actuation = ActuatorCommand(reference, value)
        self.logger.info(f"Received actuation command: {actuation}")

        return actuation

    def parse_firmware_install(self, message: Message) -> str:
        """
        Return file name from message.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            file_name = payload["fileName"]
        except Exception:
            self.logger.warning(
                f"Received invalid firmware install message: {message}"
            )
            file_name = ""

        self.logger.debug(f"File name: {file_name}")
        return file_name

    def parse_file_binary(self, message: Message) -> FileTransferPackage:
        """
        Parse the message into a file transfer package.

        :param message: The message received
        :type message: Message
        :returns: file_transfer_package
        :rtype: FileTransferPackage
        """
        if not isinstance(message.payload, bytes):
            self.logger.warning(
                f"Expected payload in bytes, got {type(message.payload)}"
            )
            self.logger.info("Treating as invalid package")
            file_transfer_package = FileTransferPackage(b"", b"", b"")
        else:
            try:
                if len(message.payload) < 65:
                    raise ValueError(
                        "Received file transfer package too small"
                    )
                previous_hash = message.payload[:32]
                data = message.payload[32 : len(message.payload) - 32]
                current_hash = message.payload[-32:]

                file_transfer_package = FileTransferPackage(
                    previous_hash, data, current_hash
                )
                self.logger.debug(
                    "Received file transfer package: "
                    f"FileTransferPackage(previous_hash={previous_hash!r},"
                    f" data={len(data)} bytes, current_hash={current_hash!r}"
                )
            except Exception:
                self.logger.warning("Received malformed file chunk!")
                file_transfer_package = FileTransferPackage(b"", b"", b"")
        return file_transfer_package

    def parse_configuration(self, message: Message) -> ConfigurationCommand:
        """
        Parse the message into a configuration command.

        :param message: The message received
        :type message: Message

        :returns: configuration
        :rtype: ConfigurationCommand
        """
        self.logger.debug(f"{message}")

        payload = json.loads(message.payload.decode("utf-8"))  # type: ignore
        values = payload["values"]

        if isinstance(values, dict):
            for reference, value in values.items():
                if value in ["true", "false"]:
                    value = bool(strtobool(value))
                if isinstance(value, str):
                    value = value.replace("\\n", "\n")
                    try:
                        if "." in value:
                            value = float(value)
                        else:
                            value = int(value)
                    except ValueError:
                        pass
                values[reference] = value

        configuration = ConfigurationCommand(values)
        self.logger.info(f"Received configuration command: {configuration}")
        return configuration

    def parse_file_delete_command(self, message: Message) -> str:
        """
        Parse the message into a file name to delete.

        :param message: The message received
        :type message: Message
        :returns: file_name
        :rtype: str
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            self.logger.debug(f'file name: {payload["fileName"]}')
            return payload["fileName"]
        except Exception:
            self.logger.warning(
                f"Failed to get file name from message {message}"
            )
            return ""

    def parse_file_url(self, message: Message) -> str:
        """
        Parse the message into a URL string.

        :param message: The message received
        :type message: Message
        :returns: file_url
        :rtype: str
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            self.logger.debug(f'file URL: {payload["fileUrl"]}')
            return payload["fileUrl"]
        except Exception:
            self.logger.warning(
                f"Failed to get file URL from message {message}"
            )
            return ""

    def parse_file_initiate(self, message: Message) -> Tuple[str, int, str]:
        """
        Return file name, file size and file hash from message.

        :param message: The message received
        :type message: Message
        :returns: (file_name, file_size, file_hash)
        :rtype: Tuple[str, int, str]
        """
        self.logger.debug(f"{message}")
        try:
            payload = json.loads(
                message.payload.decode("utf-8")  # type: ignore
            )
            self.logger.debug(
                f'file_name={payload["fileName"]}, '
                f'file_size={payload["fileSize"]}, '
                f'file_hash={payload["fileHash"]}'
            )
            return (
                payload["fileName"],
                payload["fileSize"],
                payload["fileHash"],
            )
        except Exception:
            self.logger.warning(
                f"Received invalid file upload initiate message: {message}"
            )
            return "", 0, ""



"""Create messages from data that conform to device's specified protocol."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

from wolk.model.actuator_status import ActuatorStatus
from wolk.model.alarm import Alarm
from wolk.model.file_management_status import FileManagementStatus
from wolk.model.firmware_update_status import FirmwareUpdateStatus
from wolk.model.message import Message
from wolk.model.sensor_reading import SensorReading


class MessageFactory(ABC):
    """Serialize messages to be sent to WolkAbout IoT Platform."""

    @abstractmethod
    def make_from_sensor_reading(
        self, reading: Union[SensorReading, List[SensorReading]]
    ) -> Message:
        """
        Serialize a sensor reading to be sent to WolkAbout IoT Platform.

        :param reading: Reading to be serialized
        :type reading: Union[SensorReading, List[SensorReading]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_sensor_readings(
        self,
        readings: Dict[
            str,
            Union[bool, str, int, Tuple[int, ...], float, Tuple[float, ...]],
        ],
        timestamp: Optional[int] = None,
    ) -> Message:
        """
        Serialize multiple sensor readings to send to WolkAbout IoT Platform.

        :param readings: Dictionary of reference: value pairs
        :type readings: Dict[str,Union[bool, str, int, Tuple[int, ...], float, Tuple[float, ...]]]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_alarm(self, alarm: Alarm) -> Message:
        """
        Serialize an alarm event to be sent to WolkAbout IoT Platform.

        :param alarm: Alarm to be serialized
        :type alarm: Alarm
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_actuator_status(self, actuator: ActuatorStatus) -> Message:
        """
        Serialize an actuator status to be sent to WolkAbout IoT Platform.

        :param actuator: Actuator status to be serialized
        :type actuator: ActuatorStatus
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_firmware_update_status(
        self, firmware_update_status: FirmwareUpdateStatus
    ) -> Message:
        """
        Report the current status of the firmware update process.

        :param firmware_update_status: Current status of the firmware update process
        :type firmware_update_status: FirmwareUpdateStatus
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_package_request(
        self, file_name: str, chunk_index: int, chunk_size: int
    ) -> Message:
        """
        Request a package of the file from WolkAbout IoT Platform.

        :param file_name: Name of the file that contains the requested package
        :type file_name: str
        :param chunk_index: Index of the requested package
        :type chunk_index: int
        :param chunk_size: Size of the requested package
        :type chunk_size: int
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_firmware_version_update(self, version: str) -> Message:
        """
        Report the device's current firmware version to WolkAbout IoT Platform.

        :param version: Current device firmware version
        :type version: str
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_configuration(
        self, configuration: Dict[str, Optional[Union[int, float, bool, str]]]
    ) -> Message:
        """
        Serialize device's configuration to be sent to WolkAbout IoT Platform.

        :param configuration: Device's current configuration
        :type configuration: dict
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_file_list_update(self, file_list: List[str]) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[str]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_file_list_request(self, file_list: List[str]) -> Message:
        """
        Serialize list of files present on device.

        :param file_list: Files present on device
        :type file_list: List[str]
        :returns: message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_from_file_management_status(
        self, status: FileManagementStatus, file_name: str
    ) -> Message:
        """
        Serialize device's current file management status.

        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Name of file being transfered
        :type file_name: str
        :returns: message
        :rtype: Message
        """

    @abstractmethod
    def make_from_file_url_status(
        self,
        file_url: str,
        status: FileManagementStatus,
        file_name: Optional[str] = None,
    ) -> Message:
        """
        Serialize device's current file URL download status.

        :param file_url: URL from where the file is to be downloaded
        :type file_url: str
        :param status: Current file management status
        :type status: FileManagementStatus
        :param file_name: Only present when download of file is completed
        :type file_name: Optional[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def make_last_will_message(self) -> Message:
        """
        Serialize a last will message if device disconnects unexpectedly.

        :returns: Last will message
        :rtype: Message
        """
        raise NotImplementedError()

    @abstractmethod
    def make_keep_alive_message(self) -> Message:
        """
        Serialize a keep alive message.

        :returns: keep alive message
        :rtype: Message
        """
        raise NotImplementedError()



"""Tests for WolkAboutProtocolMessageDeserializer."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import json
import logging
import sys
import unittest

sys.path.append("..")  # noqa

from wolk.model.device import Device
from wolk.model.message import Message
from wolk.model.actuator_command import ActuatorCommand
from wolk.model.configuration_command import ConfigurationCommand
from wolk.model.file_transfer_package import FileTransferPackage
from wolk.wolkabout_protocol_message_deserializer import (
    WolkAboutProtocolMessageDeserializer as WAPMD,
)


class WolkAboutProtocolMessageDeserializerTests(unittest.TestCase):
    """Tests for deserializing messages using WolkAbout Protocol."""

    device = Device(
        key="some_key",
        password="some_password",
        actuator_references=["SW", "SL", "ST"],
    )

    expected_topics = [
        WAPMD.KEEP_ALIVE_RESPONSE + device.key,
        WAPMD.CONFIGURATION_SET + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_BINARY_RESPONSE + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_DELETE + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_PURGE + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_LIST_CONFIRM + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_LIST_REQUEST + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_UPLOAD_ABORT + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_UPLOAD_INITIATE + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FILE_URL_DOWNLOAD_ABORT
        + WAPMD.DEVICE_PATH_DELIMITER
        + device.key,
        WAPMD.FILE_URL_DOWNLOAD_INITIATE
        + WAPMD.DEVICE_PATH_DELIMITER
        + device.key,
        WAPMD.FIRMWARE_UPDATE_ABORT + WAPMD.DEVICE_PATH_DELIMITER + device.key,
        WAPMD.FIRMWARE_UPDATE_INSTALL
        + WAPMD.DEVICE_PATH_DELIMITER
        + device.key,
    ]
    for reference in device.actuator_references:
        expected_topics.append(
            WAPMD.ACTUATOR_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + device.key
            + WAPMD.CHANNEL_DELIMITER
            + WAPMD.REFERENCE_PATH_PREFIX
            + reference
        )

    def test_init(self):
        """Test creating a deserializer and assert topic lists match."""
        deserializer = WAPMD(self.device)

        self.assertEqual(self.expected_topics, deserializer.inbound_topics)

    def test_get_inbound_topics(self):
        """Test inbound topics are valid for device."""
        deserializer = WAPMD(self.device)

        self.assertEqual(
            self.expected_topics, deserializer.get_inbound_topics()
        )

    def test_is_actuation_command(self):
        """Test if message is actuation command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.ACTUATOR_SET, None)

        self.assertTrue(deserializer.is_actuation_command(message))

    def test_is_keep_alive_response(self):
        """Test if message is keep alive response."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.KEEP_ALIVE_RESPONSE, None)

        self.assertTrue(deserializer.is_keep_alive_response(message))

    def test_is_firmware_install(self):
        """Test if message is firmware install command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.ACTUATOR_SET, None)

        self.assertFalse(deserializer.is_firmware_install(message))

    def test_is_firmware_abort(self):
        """Test if message is firmware abort command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FIRMWARE_UPDATE_ABORT, None)

        self.assertTrue(deserializer.is_firmware_abort(message))

    def test_is_file_binary(self):
        """Test if message is file binary."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_BINARY_RESPONSE, None)

        self.assertTrue(deserializer.is_file_binary_response(message))

    def test_is_configuration_command(self):
        """Test if message is configuration command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.CONFIGURATION_SET, None)

        self.assertTrue(deserializer.is_configuration_command(message))

    def test_is_file_delete_command(self):
        """Test if message is file delete command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_DELETE, None)

        self.assertTrue(deserializer.is_file_delete_command(message))

    def test_is_file_purge_command(self):
        """Test if message is file purge command."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_PURGE, None)

        self.assertTrue(deserializer.is_file_purge_command(message))

    def test_is_file_list_confirm(self):
        """Test if message is file list confirm."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_LIST_CONFIRM, None)

        self.assertTrue(deserializer.is_file_list_confirm(message))

    def test_is_file_list_request(self):
        """Test if message is file list request."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_LIST_REQUEST, None)

        self.assertTrue(deserializer.is_file_list_request(message))

    def test_is_file_upload_initiate(self):
        """Test if message is file upload initiate."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_UPLOAD_INITIATE, None)

        self.assertTrue(deserializer.is_file_upload_initiate(message))

    def test_is_file_upload_abort(self):
        """Test if message is file upload abort."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_UPLOAD_ABORT, None)

        self.assertTrue(deserializer.is_file_upload_abort(message))

    def test_is_file_url_initiate(self):
        """Test if message is file url initiate."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_URL_DOWNLOAD_INITIATE, None)

        self.assertTrue(deserializer.is_file_url_initiate(message))

    def test_is_file_url_abort(self):
        """Test if message is file url abort."""
        deserializer = WAPMD(self.device)
        message = Message(WAPMD.FILE_URL_DOWNLOAD_ABORT, None)

        self.assertTrue(deserializer.is_file_url_abort(message))

    def test_parse_acutator_command_set_bool(self):
        """Test parse actuator command with command set and value type bool."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)  # Disable logging
        reference = "SW"
        value = False

        incoming_topic = (
            WAPMD.ACTUATOR_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
            + WAPMD.CHANNEL_DELIMITER
            + WAPMD.REFERENCE_PATH_PREFIX
            + reference
        )
        incoming_payload = bytearray(
            json.dumps({"value": str(value).lower()}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ActuatorCommand(reference, value)

        self.assertEqual(
            expected, deserializer.parse_actuator_command(incoming_message)
        )

    def test_parse_acutator_command_set_string_with_newline(self):
        """Test parse actuator command with command set and type string."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "SW"
        value = "string\nstring"
        expected_value = value.replace("\n", "\\n")

        incoming_topic = (
            WAPMD.ACTUATOR_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
            + WAPMD.CHANNEL_DELIMITER
            + WAPMD.REFERENCE_PATH_PREFIX
            + reference
        )
        incoming_payload = bytearray(
            json.dumps({"value": expected_value}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ActuatorCommand(reference, value)

        self.assertEqual(
            expected, deserializer.parse_actuator_command(incoming_message)
        )

    def test_parse_acutator_command_set_float(self):
        """Test parse actuator command with command set and type float."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "SL"
        value = "12.3"

        incoming_topic = (
            WAPMD.ACTUATOR_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
            + WAPMD.CHANNEL_DELIMITER
            + WAPMD.REFERENCE_PATH_PREFIX
            + reference
        )
        incoming_payload = bytearray(json.dumps({"value": value}), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ActuatorCommand(reference, float(value))

        self.assertEqual(
            expected, deserializer.parse_actuator_command(incoming_message)
        )

    def test_parse_acutator_command_set_int(self):
        """Test parse actuator command with command set and type int."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "SL"
        value = "12"

        incoming_topic = (
            WAPMD.ACTUATOR_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
            + WAPMD.CHANNEL_DELIMITER
            + WAPMD.REFERENCE_PATH_PREFIX
            + reference
        )
        incoming_payload = bytearray(json.dumps({"value": value}), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ActuatorCommand(reference, int(value))

        self.assertEqual(
            expected, deserializer.parse_actuator_command(incoming_message)
        )

    def test_parse_keep_alive_response(self):
        """Test parse keep alive response message."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        timestamp = 123

        incoming_topic = WAPMD.KEEP_ALIVE_RESPONSE + self.device.key
        incoming_payload = bytearray(json.dumps({"value": timestamp}), "utf-8")
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = timestamp

        self.assertEqual(
            expected, deserializer.parse_keep_alive_response(incoming_message)
        )

    def test_parse_firmware_install(self):
        """Test parse firmware install command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "install_me.bin"

        incoming_topic = (
            WAPMD.FIRMWARE_UPDATE_INSTALL
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"fileName": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = file_name

        self.assertEqual(
            expected, deserializer.parse_firmware_install(incoming_message)
        )

    def test_parse_firmware_install_invalid(self):
        """Test parse invalid firmware install command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "install_me.bin"

        incoming_topic = (
            WAPMD.FIRMWARE_UPDATE_INSTALL
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"file_name": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ""

        self.assertEqual(
            expected, deserializer.parse_firmware_install(incoming_message)
        )

    def test_parse_file_binary_invalid(self):
        """Test parse invalid file binary payload."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "install_me.bin"

        incoming_topic = (
            WAPMD.FILE_BINARY_RESPONSE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"file_name": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(b"", b"", b"")

        self.assertEqual(
            expected, deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_file_binary_invalid_size(self):
        """Test parse invalid file binary payload size."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        previous_hash = 15 * b"\x00"
        data = b"\x00"
        current_hash = 15 * b"\x00"

        incoming_topic = (
            WAPMD.FILE_BINARY_RESPONSE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = previous_hash + data + current_hash
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(b"", b"", b"")

        self.assertEqual(
            expected, deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_file_binary(self):
        """Test parse file binary."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        previous_hash = 32 * b"\x00"
        data = 32 * b"\x00"
        current_hash = 32 * b"\x00"

        incoming_topic = (
            WAPMD.FILE_BINARY_RESPONSE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = previous_hash + data + current_hash
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = FileTransferPackage(previous_hash, data, current_hash)

        self.assertEqual(
            expected, deserializer.parse_file_binary(incoming_message)
        )

    def test_parse_configuration_set_bool(self):
        """Test parse configuration set command with bool type."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "B"
        value = "true"

        incoming_topic = (
            WAPMD.CONFIGURATION_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"values": {reference: value}}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ConfigurationCommand({reference: True})

        self.assertEqual(
            expected, deserializer.parse_configuration(incoming_message)
        )

    def test_parse_configuration_set_string_with_newline(self):
        """Test parse configuration set command for string with newline."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "S"
        value = "string\nstring"  # escaped in json.dumps

        incoming_topic = (
            WAPMD.CONFIGURATION_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"values": {reference: value}}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ConfigurationCommand({reference: value})

        self.assertEqual(
            expected, deserializer.parse_configuration(incoming_message)
        )

    def test_parse_configuration_set_float(self):
        """Test parse configuration set command for float."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "F"
        value = 12.3

        incoming_topic = (
            WAPMD.CONFIGURATION_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"values": {reference: str(value)}}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ConfigurationCommand({reference: value})

        self.assertEqual(
            expected, deserializer.parse_configuration(incoming_message)
        )

    def test_parse_configuration_set_int(self):
        """Test parse configuration set command for int."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        reference = "F"
        value = 12

        incoming_topic = (
            WAPMD.CONFIGURATION_SET
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"values": {reference: str(value)}}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        expected = ConfigurationCommand({reference: value})

        self.assertEqual(
            expected, deserializer.parse_configuration(incoming_message)
        )

    def test_parse_file_delete_command(self):
        """Test parse file delete command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "delete_me.bin"
        expected = file_name

        incoming_topic = (
            WAPMD.FILE_DELETE + WAPMD.DEVICE_PATH_DELIMITER + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"fileName": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_delete_command(incoming_message)
        )

    def test_parse_file_delete_command_invalid(self):
        """Test parse file delete invalid command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "delete_me.bin"
        expected = ""

        incoming_topic = (
            WAPMD.FILE_DELETE + WAPMD.DEVICE_PATH_DELIMITER + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"file_name": file_name}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_delete_command(incoming_message)
        )

    def test_parse_file_url(self):
        """Test parse file url command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_url = "http://hello.there.hi/resource.png"
        expected = file_url

        incoming_topic = (
            WAPMD.FILE_URL_DOWNLOAD_INITIATE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"fileUrl": file_url}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_url(incoming_message)
        )

    def test_parse_file_url_invalid(self):
        """Test parse file url invalid command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_url = "http://hello.there.hi/resource.png"
        expected = ""

        incoming_topic = (
            WAPMD.FILE_URL_DOWNLOAD_INITIATE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps({"file_url": file_url}), "utf-8"
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_url(incoming_message)
        )

    def test_parse_file_initiate(self):
        """Test parse file initiate command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "file.bin"
        file_size = 128
        file_hash = "some_hash"
        expected = (file_name, file_size, file_hash)

        incoming_topic = (
            WAPMD.FILE_UPLOAD_INITIATE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps(
                {
                    "fileName": file_name,
                    "fileSize": file_size,
                    "fileHash": file_hash,
                }
            ),
            "utf-8",
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_initiate(incoming_message)
        )

    def test_parse_file_initiate_invalid(self):
        """Test parse file initiate invalid command."""
        deserializer = WAPMD(self.device)
        deserializer.logger.setLevel(logging.CRITICAL)
        file_name = "file.bin"
        file_size = 128
        file_hash = "some_hash"
        expected = ("", 0, "")

        incoming_topic = (
            WAPMD.FILE_UPLOAD_INITIATE
            + WAPMD.DEVICE_PATH_DELIMITER
            + self.device.key
        )
        incoming_payload = bytearray(
            json.dumps(
                {
                    "obviously_wrong_name": file_name,
                    "fileSize": file_size,
                    "fileHash": file_hash,
                }
            ),
            "utf-8",
        )
        incoming_message = Message(incoming_topic, incoming_payload)

        self.assertEqual(
            expected, deserializer.parse_file_initiate(incoming_message)
        )



"""Module responsible for handling files and file transfer."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import List
from typing import Optional

from wolk.model.file_management_status import FileManagementStatus
from wolk.model.file_transfer_package import FileTransferPackage


class FileManagement(ABC):
    """
    File transfer manager.

    Enables device to transfer files from WolkAbout IoT Platform
     package by package or/and URL download.
    """

    @abstractmethod
    def __init__(
        self,
        status_callback: Callable[[str, FileManagementStatus], None],
        packet_request_callback: Callable[[str, int, int], None],
        url_status_callback: Callable[
            [str, FileManagementStatus, Optional[str]], None
        ],
    ) -> None:
        """
        Enable file management for device.

        :param status_callback: Reporting current file transfer status
        :type status_callback: Callable[[FileManagementStatus], None]
        :param packet_request_callback: Request file packet from Platform
        :type packet_request_callback: Callable[[str, int, int], None]
        :param url_status_callback: Report file url download status
        :type url_status_callback: Callable[[str, FileManagementStatus, Optional[str]], None]
        """
        raise NotImplementedError()

    @abstractmethod
    def configure(
        self,
        preferred_package_size: int,
        max_file_size: int,
        file_directory: str,
    ) -> None:
        """
        Configure options for file management module.

        :param preferred_package_size: Size in bytes
        :type preferred_package_size: int
        :param max_file_size: Maximum file size that can be stored
        :type max_file_size: int
        :param file_directory: Path to where files are stored
        :type file_directory: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_upload_initiation(
        self, file_name: str, file_size: int, file_hash: str
    ) -> None:
        """
        Start making package requests and set status to file transfer.

        :param file_name: File name
        :type file_name: str
        :param file_size: Size in bytes
        :type file_size: int
        :param file_hash: base64 encoded sha256 hash of file
        :type file_hash: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_upload_abort(self) -> None:
        """Abort file upload and revert to idle status."""
        raise NotImplementedError()

    @abstractmethod
    def handle_file_binary_response(
        self, package: FileTransferPackage
    ) -> None:
        """
        Validate received package and store or use callback to request again.

        :param package: Package of file being transfered.
        :type package: FileTransferPackage
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_url_download_initiation(self, file_url: str) -> None:
        """
        Start file transfer from specified URL.

        :param file_url: URL from where to download file
        :type file_url: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_url_download_abort(self) -> None:
        """Abort file URL download."""
        raise NotImplementedError()

    @abstractmethod
    def get_file_list(self) -> List[str]:
        """
        Return list of files present on device.

        :returns: file_list
        :rtype: List[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def get_file_path(self, file_name: str) -> Optional[str]:
        """
        Return path to file if it exists.

        :param file_name: File for which to get path
        :type file_name: str
        :returns: file_path
        :rtype: Optional[str]
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_list_confirm(self) -> None:
        """Acknowledge file list response from WolkAbout IoT Platform."""
        raise NotImplementedError()

    @abstractmethod
    def handle_file_delete(self, file_name: str) -> None:
        """
        Delete file from device.

        :param file_name: File to be deleted
        :type file_name: str
        """
        raise NotImplementedError()

    @abstractmethod
    def handle_file_purge(self) -> None:
        """Delete all files from device."""
        raise NotImplementedError()



"""Tests for MessageDeque."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import logging
import sys
import unittest

sys.path.append("..")  # noqa

from wolk.message_deque import MessageDeque
from wolk.model.message import Message


class TestMessageDeque(unittest.TestCase):
    """Tests for MessageDeque."""

    default_message = Message("some_topic", "some_payload")

    def test_init(self):
        """Test creating an instance of MessageDeque."""
        message_deque = MessageDeque()
        self.assertIsNotNone(message_deque.queue)

    def test_put_no_message(self):
        """Test passing None to put method."""
        message_deque = MessageDeque()
        message_deque.logger.setLevel(logging.CRITICAL)  # Disable logging
        self.assertFalse(message_deque.put(None))

    def test_put_message(self):
        """Test passing None to put method."""
        message_deque = MessageDeque()
        self.assertTrue(message_deque.put(self.default_message))

    def test_get_message(self):
        """Test getting message from queue."""
        message_deque = MessageDeque()
        message_deque.put(self.default_message)
        self.assertEqual(self.default_message, message_deque.get())

    def test_get_empty(self):
        """Test getting no message from queue."""
        message_deque = MessageDeque()
        self.assertIsNone(message_deque.get())

    def test_peek_empty(self):
        """Test getting no message from queue peek."""
        message_deque = MessageDeque()
        self.assertIsNone(message_deque.peek())

    def test_peek_not_empty(self):
        """Test getting no message from queue peek."""
        message_deque = MessageDeque()
        message_deque.put(self.default_message)
        self.assertIsNotNone(message_deque.peek())



"""States of actuators & configurations as defined on WolkAbout IoT Platform."""
#   Copyright 2020 WolkAbout Technology s.r.o.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from enum import Enum
from enum import unique


@unique
class State(Enum):
    """
    Enumeration of available states.

    :ivar BUSY: Currently in busy state
    :vartype BUSY: str
    :ivar ERROR: Currently in error state
    :vartype ERROR: str
    :ivar READY: Currently in ready state
    :vartype READY: str
    """

    READY = "READY"
    BUSY = "BUSY"
    ERROR = "ERROR"
